#
#
# Author: Donald Beaudry <donb@sgi.com>
# Date:   March 26, 1997
#
# DESCRIPTION:
#
# This module exports a pgen grammar object that knows how to parse
# strings from conforming to the Python grammar.  The grammar object
# is pre-configured with a set of reduction functions designed to make
# processing of resulting AST much more straightforward.  
#
# USAGE:
#   import python
#   ast = python.parse_string("a=b+c", python.NT.single_input)
# 
#
# REDUCTIONS:
#
# The AST produced by the reduction functions is summerized below.
# The notation is intended to describe the tree return by a reduction.
#  XXX 
# 
# The following three start symbols are the only ones supported.  You
# might be able to get some use from any of the other non-terminals
# defined by the grammar... just don't count on it ;)
#
#
# single_input --> (NT.single_input, stmt*)
#
# file_input --> (NT.file_input, stmt*)
#
# eval_input --> (NT.eval_input, test*)
#
# stmt --> (NT.suite, stmt+)
#       |  (NT.simple_stmt, stmt, stmt+)
#       |  (NT.expr_stmt, testlist+)
#       |  (NT.print_stmt, test*, COMMA?)
#       |  (NT.del_stmt, expr*)
#       |  (NT.pass_stmt,)
#       |  (NT.break_stmt,)
#       |  (NT.continue_stmt,)
#       |  (NT.return_stmt, test*)
#       |  (NT.raise_stmt, test)
#       |  (NT.raise_stmt, test, test)
#       |  (NT.raise_stmt, test, test, test)
#       |  (NT.import_stmt, KW.import_, dotted_name+)
#       |  (NT.import_stmt, KW.from_, dotted_name, T.STAR)
#       |  (NT.import_stmt, KW.from_, dotted_name, name+)
#       |  (NT.global_stmt, name+)
#       |  (NT.exec_stmt, expr)
#       |  (NT.exec_stmt, expr, test)
#       |  (NT.exec_stmt, expr, test, test)
#       |  (NT.while_stmt, test, stmt)
#       |  (NT.while_stmt, test, stmt, stmt)
#       |  (NT.for_stmt, exprlist, testlist, stmt)
#       |  (NT.for_stmt, exprlist, testlist, stmt, stmt)
#       |  (NT.if_stmt, if_clause+, else_clause?)
#       |  (NT.try_stmt, stmt, except_clause+, else_clause?)
#       |  (NT.try_stmt, stmt, finally_clause)
#       |  (NT.funcdef, parameters, stmt)
#       |  (NT.classdef, name, testlist, stmt)
#
# parameters --> (NT.parameters, param*)
#
# param --> paramdef
#        | (T.EQUAL, paramdef, test)
#        | (T.STAR, name)
#        | (T.DOUBLESTAR, name)
# 
# paramdef --> name
#           | (T.LPAR paramdef*)
#
# dotted_name --> name
#              |  (T.DOT, name, name+)
#
# except_clause --> (KW.except_, stmt)
#                |  (KW.except_, (NT.except_clause, test), stmt)
#                |  (KW.except_, (NT.except_clause, test, test), stmt)
#
# finally_clause --> (KW.finally_, stmt)
#
# if_clause --> (KW.if_, stmt)
#
# else_clause --> (KW.else_, stmt)
#
# exprlist --> (NT.testlist, expr+)
#
# testlist --> (NT.testlist, test+)
#
# exprlist --> (NT.exprlist, expr+)
#
# test --> (KW.or, test, test+)
#       |  (KW.and, test, test+)
#       |  (KW.not, test)
#       |  (T.LESS, test, test)
#       |  (T.GREATER, test, test)
#       |  (T.EQEQUAL, test, test)
#       |  (T.GREATEREQUAL, test, test)
#       |  (T.LESSEQUAL, test, test)
#       |  (T.NOTEQUAL, test, test)
#       |  (KW.in, test, test)
#       |  (NOT_IN, test, test)
#       |  (KW.is, test, test)
#       |  (IS_NOT, test, test)
#       |  (NT.lambdef, parameters, test)
#       |  expr
#
# expr --> (T.VBAR, expr, expr+)
#       |  (T.CIRCUMFLEX, expr, expr+)
#       |  (T.AMPER, expr, expr+)
#       |  (T.LEFTSHIFT, expr, expr+)
#       |  (T.RIGHTSHIFT, expr, expr+)
#       |  (T.PLUS, expr+)
#       |  (T.MINUS, expr+)
#       |  (T.STAR, expr, expr+)
#       |  (T.SLASH, expr, expr+)
#       |  (T.PERCENT, expr, expr+)
#       |  (T.TILDE, expr)
#       |  (T.DOUBLESTAR, expr, expr)
#       |  (CALL, expr, argument+)
#       |  (NT.subscript, expr, subscript+)
#       |  (T.DOT, expr, name+)
#       |  (T.NAME, name)
#       |  (T.NUMBER, number)
#       |  (T.STRING, string+)
#       |  (T.LPAR, expr*)
#       |  (T.LSQB, expr*)
#       |  ('`', expr*)
#       |  (T.LBRACE, (T.COLON, expr, expr)*)
#
# subscript --> test
#            |  (ELIPSE,)
#            |  (T.sliceop, test)
#            |  (T.sliceop, test, test)
#            |  (T.slice, test, test, test)
#
# argument --> test
#           |  (T.EQUAL, name, test)
#

#
# This is the file Grammar/Grammar from Python-1.4
# 
GRAMMAR = """
# Grammar for Python

# Commands for Kees Blom's railroad program
#diagram:token NAME
#diagram:token NUMBER
#diagram:token STRING
#diagram:token NEWLINE
#diagram:token ENDMARKER
#diagram:token INDENT
#diagram:output\input python.bla
#diagram:token DEDENT
#diagram:output\textwidth 20.04cm\oddsidemargin  0.0cm\evensidemargin 0.0cm
#diagram:rules

# Start symbols for the grammar:
#	single_input is a single interactive statement;
#	file_input is a module or sequence of commands read from an input file;
#	eval_input is the input for the eval() and input() functions.
# NB: compound_stmt in single_input is followed by extra NEWLINE!
single_input: NEWLINE | simple_stmt | compound_stmt NEWLINE
file_input: (NEWLINE | stmt)* ENDMARKER
eval_input: testlist NEWLINE* ENDMARKER

funcdef: 'def' NAME parameters ':' suite
parameters: '(' [varargslist] ')'
varargslist: (fpdef ['=' test] ',')* ('*' NAME [',' ('**'|'*' '*') NAME] | ('**'|'*' '*') NAME) | fpdef ['=' test] (',' fpdef ['=' test])* [',']
fpdef: NAME | '(' fplist ')'
fplist: fpdef (',' fpdef)* [',']

stmt: simple_stmt | compound_stmt
simple_stmt: small_stmt (';' small_stmt)* [';'] NEWLINE
#small_stmt: expr_stmt | print_stmt  | del_stmt | pass_stmt | flow_stmt | import_stmt | global_stmt | access_stmt | exec_stmt
small_stmt: expr_stmt | print_stmt  | del_stmt | pass_stmt | flow_stmt | import_stmt | global_stmt | exec_stmt
expr_stmt: testlist ('=' testlist)*
# For assignments, additional restrictions enforced by the interpreter
print_stmt: 'print' (test ',')* [test]
del_stmt: 'del' exprlist
pass_stmt: 'pass'
flow_stmt: break_stmt | continue_stmt | return_stmt | raise_stmt
break_stmt: 'break'
continue_stmt: 'continue'
return_stmt: 'return' [testlist]
raise_stmt: 'raise' test [',' test [',' test]]
import_stmt: 'import' dotted_name (',' dotted_name)* | 'from' dotted_name 'import' ('*' | NAME (',' NAME)*)
dotted_name: NAME ('.' NAME)*
global_stmt: 'global' NAME (',' NAME)*
#access_stmt: 'access' ('*' | NAME (',' NAME)*) ':' accesstype  (',' accesstype)*
#accesstype: NAME+
## accesstype should be ('public' | 'protected' | 'private') ['read'] ['write']
## but can't be because that would create undesirable reserved words!
exec_stmt: 'exec' expr ['in' test [',' test]]

compound_stmt: if_stmt | while_stmt | for_stmt | try_stmt | funcdef | classdef
if_stmt: 'if' test ':' suite ('elif' test ':' suite)* ['else' ':' suite]
while_stmt: 'while' test ':' suite ['else' ':' suite]
for_stmt: 'for' exprlist 'in' testlist ':' suite ['else' ':' suite]
try_stmt: ('try' ':' suite (except_clause ':' suite)+ #diagram:break
           ['else' ':' suite] | 'try' ':' suite 'finally' ':' suite)
# NB compile.c makes sure that the default except clause is last
except_clause: 'except' [test [',' test]]
suite: simple_stmt | NEWLINE INDENT stmt+ DEDENT

test: and_test ('or' and_test)* | lambdef
and_test: not_test ('and' not_test)*
not_test: 'not' not_test | comparison
comparison: expr (comp_op expr)*
comp_op: '<'|'>'|'=='|'>='|'<='|'<>'|'!='|'in'|'not' 'in'|'is'|'is' 'not'
expr: xor_expr ('|' xor_expr)*
xor_expr: and_expr ('^' and_expr)*
and_expr: shift_expr ('&' shift_expr)*
shift_expr: arith_expr (('<<'|'>>') arith_expr)*
arith_expr: term (('+'|'-') term)*
term: factor (('*'|'/'|'%') factor)*
factor: ('+'|'-'|'~') factor | power
power: atom trailer* ('**' factor)*
atom: '(' [testlist] ')' | '[' [testlist] ']' | '{' [dictmaker] '}' | '`' testlist '`' | NAME | NUMBER | STRING+
lambdef: 'lambda' [varargslist] ':' test
trailer: '(' [arglist] ')' | '[' subscriptlist ']' | '.' NAME
subscriptlist: subscript (',' subscript)* [',']
subscript: '.' '.' '.' | test | [test] ':' [test] [sliceop]
sliceop: ':' [test]
exprlist: expr (',' expr)* [',']
testlist: test (',' test)* [',']
dictmaker: test ':' test (',' test ':' test)* [',']

classdef: 'class' NAME ['(' testlist ')'] ':' suite

arglist:  argument (',' argument)* [',']
argument: [test '='] test	# Really [keyword '='] test

"""

import pgen, sys, string, os, operator

CALL = "call"
ELIPSE = "..."
KWERR = "*kwerr*"
NOT_IN = "not_in"
IS_NOT = "is_not"


def keep(children, *keep):
    return filter(lambda x, keep=keep: x[0] in keep, children)

def strip(children):
    return tuple(map(lambda x: x[1], children))

def to_prefix(s):
    l = len(s)
    if l == 0:
        return ()
    lop=''
    t1 = s[0][1]
    i = 1
    while i < l:
        op = s[i][0]
        t2 = s[i+1][1]
        if lop != op:
            t1 = (op, t1, t2)
            lop = op
        else:
            t1 = t1 + (t2,)
        i = i + 2
    return t1


#
# The reductions class is used only for name space control.  It is
# never instantiated.
#
class reductions:
    #
    # single_input --> (NT.single_input,)
    #               |  (NT.single_input, stmt)
    #
    # single_input: NEWLINE | simple_stmt | compound_stmt NEWLINE
    def single_input(g, tok, where, child):
        child = strip(keep(child, g.NT.simple_stmt, g.NT.compound_stmt))
        if len(child):
            return (tok, child[0])
        else:
            return (tok,)

    #
    # file_input --> (NT.file_input, stmt*)
    #
    # file_input: (NEWLINE | stmt)* ENDMARKER
    def file_input(g, tok, where, child):
        child = strip(keep(child, g.NT.stmt))
        if len(child):
            return (tok,) + child
        else:
            return (tok,)


    #
    # eval_input --> (NT.eval_input, test*)
    #
    # eval_input: testlist NEWLINE* ENDMARKER
    def eval_input(g, tok, where, child):
        child = child[0][1][1:]
        if len(child):
            return (tok,) + child
        else:
            return (tok,)

    #
    #
    # funcdef --> ('funcdef', parameters, stmt)
    #
    # funcdef: 'def' NAME parameters ':' suite
    def funcdef(g, tok, where, child):
        return (tok,) + (child[1][1], child[2], child[4][1])

    #
    # parameters --> (NT.parameters, param*)
    #
    # parameters: '(' [varargslist] ')'
    def parameters(g, tok, where, child):
        if len(child) == 3:
            return (tok,) + child[1][1]
        else:
            return (tok,)

    #
    # varargslist --> (T.varargslist, (param*))
    #
    # param --> fpdef
    #        | (T.EQUAL, fpdef, test)
    #        | (T.STAR, name)
    #        | (T.DOUBLESTAR, name)
    #
    # varargslist: (fpdef ['=' test] ',')* ('*' NAME [',' ('**'|'*' '*') NAME] 
    #                                       | ('**'|'*' '*') NAME)
    #              | fpdef ['=' test] (',' fpdef ['=' test])* [',']
    def varargslist(g, tok, where, child):
        r = []
        add = r.append
        i = 0
        nch = len(child)
        while i < nch:
            ch = child[i]
            if ch[0] is g.T.COMMA:
                i = i + 1
            elif ch[0] is g.T.STAR:
                ch = child[i+1]
                if ch[0] is g.T.STAR:
                    ch = child[i+2]
                    add(g.T.DOUBLESTAR, ch[1])
                    i = i + 3
                else:
                    add(g.T.STAR, ch[1])
                    i = i + 2
            elif ch[0] is g.T.DOUBLESTAR:
                ch = child[i+1]
                add(g.T.DOUBLESTAR, ch[1])
                i = i + 2
            elif i+1 < nch and child[i+1][0] is g.T.EQUAL:
                add(g.T.EQUAL, ch[1], child[i+2][1])
                i = i + 3
            else:
                add(ch[1])
                i = i + 1
        return (tok, tuple(r))

    # 
    # fpdef --> name
    #        | (T.LPAR fpdef*)
    #
    # fpdef: NAME | '(' fplist ')'
    def fpdef(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, (child[0][0],) + child[1][1])

    # fplist: fpdef (',' fpdef)* [',']
    def fplist(g, tok, where, child):
        return (tok, strip(keep(child, g.NT.fpdef)))

    #
    # stmt --> simple_stmt
    #       |  compound_stmt
    #
    # stmt: simple_stmt | compound_stmt
    def stmt(g, tok, where, child):
        return (tok, child[0][1])

    #
    # simple_stmt --> small_stmt
    #              |  (NT.simple_stmt, small_stmt*)
    #
    # simple_stmt: small_stmt (';' small_stmt)* [';'] NEWLINE
    def simple_stmt(g, tok, where, child):
        child = strip(keep(child, g.NT.small_stmt))
        if len(child) == 1:
            return (tok, child[0])
        else:
            return (tok, (tok,) + child)

    #
    # small_stmt --> expr_stmt
    #             |  print_stmt
    #             |  del_stmt
    #             |  pass_stmt
    #             |  import_stmt
    #             |  global_stmt
    #             |  exec_stmt
    #             |  flow_stmt
    #
    # small_stmt: expr_stmt | print_stmt | del_stmt
    #            | pass_stmt | flow_stmt | import_stmt | global_stmt | exec_stmt
    def small_stmt(g, tok, where, child):
        return (tok, child[0])

    #
    # expr_stmt --> (NT.expr_stmt, testlist+)
    #
    # expr_stmt: testlist ('=' testlist)*
    # # For assignments, additional restrictions enforced by the interpreter
    def expr_stmt(g, tok, where, child):
        child = strip(keep(child, g.NT.testlist))
        return (tok,) + child

    #
    # print_stmt --> (NT.print_stmt, test*, COMMA?)
    #
    # print_stmt: 'print' (test ',')* [test]
    def print_stmt(g, tok, where, child):
        if child[-1][0] is g.T.COMMA:
            return (tok,) + strip(keep(child, g.NT.test)) + (g.T.COMMA,)
        else:
            return (tok,) + strip(keep(child, g.NT.test))

    #
    # del_stmt --> (NT.del_stmt, expr*)
    #
    # del_stmt: 'del' exprlist
    def del_stmt(g, tok, where, child):
        return (tok,) + child[1][1][1:]

    #
    # pass_stmt --> (NT.pass_stmt,)
    #
    # pass_stmt: 'pass'
    def pass_stmt(g, tok, where, child):
        return (tok,)

    #
    # flow_stmt --> break_stmt
    #            |  continue_stmt
    #            |  return_stmt
    #            |  raise_stmt
    #
    # flow_stmt: break_stmt | continue_stmt | return_stmt | raise_stmt
    def flow_stmt(g, tok, where, child):
        return child[0]

    #
    # break_stmt --> (NT.break_stmt,)
    #
    # break_stmt: 'break'
    def break_stmt(g, tok, where, child):
        return (tok,)

    #
    # continue_stmt --> (NT.continue_stmt,)
    #
    # continue_stmt: 'continue'
    def continue_stmt(g, tok, where, child):
        return (tok,)

    #
    # return_stmt --> (NT.return_stmt, test*)
    #
    # return_stmt: 'return' [testlist]
    def return_stmt(g, tok, where, child):
        if len(child) == 2:
            return (tok,) + child[1][1:]
        else:
            return (tok,)

    #
    # raise_stmt --> (NT.raise_stmt, test)
    #             |  (NT.raise_stmt, test, test)
    #             |  (NT.raise_stmt, test, test, test)
    #
    # raise_stmt: 'raise' test [',' test [',' test]]
    def raise_stmt(g, tok, where, child):
        return (tok,) + strip(keep(child, g.NT.test))

    #
    # import_stmt --> (NT.import_stmt, KW.import_, dotted_name+)
    #              |  (NT.import_stmt, KW.from_, dotted_name, T.STAR)
    #              |  (NT.import_stmt, KW.from, dotted_name, name+)
    #
    #
    # import_stmt: 'import' dotted_name (',' dotted_name)*
    #              | 'from' dotted_name 'import' ('*' | NAME (',' NAME)*)
    def import_stmt(g, tok, where, child):
        return (tok, child[0][1]) \
               + strip(keep(child, g.NT.dotted_name, g.T.STAR, g.T.NAME))

    #
    # dotted_name --> name
    #              |  (NT.dotted_name, name, name+)
    #
    # dotted_name: NAME ('.' NAME)*
    def dotted_name(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, (g.T.DOT,) + strip(keep(child, g.T.NAME)))

    #
    # global_stmt --> (NT.global_stmt, name+)
    #
    # global_stmt: 'global' NAME (',' NAME)*
    def global_stmt(g, tok, where, child):
        return (tok,) + strip(keep(child, g.T.NAME))

    #
    # exec_stmt --> (NT.exec_stmt, expr)
    #            |  (NT.exec_stmt, expr, test)
    #            |  (NT.exec_stmt, expr, test, test)
    #
    # exec_stmt: 'exec' expr ['in' test [',' test]]
    def exec_stmt(g, tok, where, child):
        return (tok,) + strip(keep(child, g.NT.expr, g.NT.test))


    #
    # compound_stmt --> if_stmt
    #                |  while_stmt
    #                |  for_stmt
    #                |  try_stmt
    #                |  funcdef
    #                |  classdef
    #
    # compound_stmt: if_stmt | while_stmt | for_stmt
    #                | try_stmt | funcdef | classdef
    def compound_stmt(g, tok, where, child):
        return (tok, child[0])

    #
    # if_stmt --> (NT.if_stmt, (KW.if_, test, stmt)+, (KW.else_, stmt)?)
    #
    # if_stmt: 'if' test ':' suite ('elif' test ':' suite)* ['else' ':' suite]
    def if_stmt(g, tok, where, child):
        r = []
        add = r.append
        add(tok)
        nch = len(child)
        if_kws = [g.KW.elif_, g.KW.if_]
        i = 0
        while i < nch:
            ch = child[i]
            if ch[0] in if_kws:
                add(g.KW.if_, child[i+1][1], child[i+3][1])
                i = i + 4
            else:
                add(g.KW.else_, child[i+2][1])
                i = i + 3
        return tuple(r)


    #
    # while_stmt--> (NT.while_stmt, test, stmt, stmt?)
    #
    # while_stmt: 'while' test ':' suite ['else' ':' suite]
    def while_stmt(g, tok, where, child):
        return (tok,) + strip(keep(child, g.NT.test, g.NT.suite))


    #
    # for_stmt --> (NT.for_stmt, exprlist, testlist, stmt)
    #           |  (NT.for_stmt, exprlist, testlist, stmt, stmt)
    #
    # for_stmt: 'for' exprlist 'in' testlist ':' suite ['else' ':' suite]
    def for_stmt(g, tok, where, child):
        child = keep(child, g.NT.exprlist, g.NT.testlist, g.NT.suite)
        if len(child) == 3:
            return (tok, child[0][1], child[1][1], child[2][1])
        else:
            return (tok, child[0][1], child[1][1], child[2][1], child[-1][1])
    #
    # try_stmt --> (NT.try_stmt, stmt, {except_clause, stmt}+, stmt?)
    #            | (NT.try_stmt, stmt, KW.finally_, stmt)
    #
    # try_stmt: ('try' ':' suite (except_clause ':' suite)+ #diagram:break
    #            ['else' ':' suite] | 'try' ':' suite 'finally' ':' suite)
    def try_stmt(g, tok, where, child):
        if child[3][0] is g.KW.finally_:
            return (tok,) + (child[2][1], g.KW.finally_, child[5][1])
        else:
            return (tok,) + strip(keep(child, g.NT.except_clause, g.NT.suite))

    #
    # except_clause --> (NT.except_clause,)
    #                |  (NT.except_clause, test)
    #                |  (NT.except_clause, test, test)
    #
    # # NB compile.c makes sure that the default except clause is last
    # except_clause: 'except' [test [',' test]]
    def except_clause(g, tok, where, child):
        return (tok, (tok,) + strip(keep(child, g.NT.test)))

    #
    # suite --> simple_stmt
    #        |  (NT.suite, stmt+)
    #
    # suite: simple_stmt | NEWLINE INDENT stmt+ DEDENT
    def suite(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            child = strip(child[2:-1])
            return (tok, (tok,) + child)

    #
    # test --> (KW.or, test, test+)
    #       |  (KW.and, test, test+)
    #       |  (KW.not, test)
    #       |  (T.LESS, test, test)
    #       |  (T.GREATER, test, test)
    #       |  (T.EQEQUAL, test, test)
    #       |  (T.GREATEREQUAL, test, test)
    #       |  (T.LESSEQUAL, test, test)
    #       |  (T.NOTEQUAL, test, test)
    #       |  (KW.in, test, test)
    #       |  (NOT_IN, test, test)
    #       |  (KW.is, test, test)
    #       |  (IS_NOT, test, test)
    #       |  (NT.lambdef, parameters, test)
    #       |  expr
    #
    # test: and_test ('or' and_test)* | lambdef
    def test(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, (child[1][0],) + strip(keep(child, g.NT.and_test)))

    #    
    # and_test --> not_test
    #           | (KW.and, not_test, not_test+)
    #
    # and_test: not_test ('and' not_test)*
    def and_test(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, (child[1][0],) + strip(keep(child, g.NT.not_test)))

    #
    # not_test --> comparison
    #           |  (KW.not, not_test)
    #
    # not_test: 'not' not_test | comparison
    def not_test(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, (child[0][0], child[1][1]))

    #
    # comparison --> expr
    #             |  (T.LESS, comparison, comparison)
    #             |  (T.GREATER, comparison, comparison)
    #             |  (T.EQEQUAL, comparison, comparison)
    #             |  (T.GREATEREQUAL, comparison, comparison)
    #             |  (T.LESSEQUAL, comparison, comparison)
    #             |  (T.NOTEQUAL, comparison, comparison)
    #             |  (KW.in, comparison, comparison)
    #             |  (NOT_IN, comparison, comparison)
    #             |  (KW.is, comparison, comparison)
    #             |  (IS_NOT, comparison, comparison)
    #
    # comparison: expr (comp_op expr)*
    def comparison(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, to_prefix(child))

    # comp_op: '<'|'>'|'=='|'>='|'<='|'<>'|'!='|'in'|'not' 'in'|'is'|'is' 'not'
    def comp_op(g, tok, where, child):
        if len(child) == 1:
            return child[0]
        elif child[0][0] == g.KW.not_:
            return (NOT_IN, 'not in') 
        else:
            return (IS_NOT, 'is not')

    #
    # expr --> (T.VBAR, expr, expr+)
    #       |  (T.CIRCUMFLEX, expr, expr+)
    #       |  (T.AMPER, expr, expr+)
    #       |  (T.LEFTSHIFT, expr, expr)
    #       |  (T.RIGHTSHIFT, expr, expr)
    #       |  (T.PLUS, expr+)
    #       |  (T.MINUS, expr+)
    #       |  (T.STAR, expr, expr)
    #       |  (T.SLASH, expr, expr)
    #       |  (T.PERCENT, expr, expr)
    #       |  (T.TILDE, expr)
    #       |  (T.DOUBLESTAR, expr, expr)
    #       |  (CALL, expr, argument+)
    #       |  (NT.subscript, expr, subscript+)
    #       |  (T.DOT, expr, name)
    #       |  (T.NAME, name)
    #       |  (T.NUMBER, number)
    #       |  (T.STRING, string+)
    #       |  (T.LPAR, expr*)
    #       |  (T.LSQB, expr*)
    #       |  ('`', expr*)
    #       |  (T.LBRACE, (T.COLON, expr, expr)*)
    #
    # expr: xor_expr ('|' xor_expr)*
    def expr(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, (child[1][0],) + strip(keep(child, g.NT.xor_expr)))

    # xor_expr --> and_expr
    #           |  (T.CIRCUMFLEX, and_expr+)
    #
    # xor_expr: and_expr ('^' and_expr)*
    def xor_expr(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, (child[1][0],) + strip(keep(child, g.NT.and_expr)))

    #
    # and_expr --> shift_expr
    #           |  (T.AMPER, shift_expr+)
    #
    # and_expr: shift_expr ('&' shift_expr)*
    def and_expr(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, (child[1][0],) + strip(keep(child, g.NT.shift_expr)))

    #    
    # shift_expr --> arith_expr
    #           |  (T.LEFTSHIFT, arith_expr, shift_expr)
    #           |  (T.RIGHTSHIFT, arith_expr, shift_expr)
    #
    # shift_expr: arith_expr (('<<'|'>>') arith_expr)*
    def shift_expr(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, to_prefix(child))

    # arith_expr --> term
    #             |  (T.PLUS, term, arith_expr)
    #             |  (T.MINUS, term, arith_expr)
    #
    # arith_expr: term (('+'|'-') term)*
    def arith_expr(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, to_prefix(child))

    # term --> factor
    #       |  (T.STAR, factor, term)
    #       |  (T.SLASH, factor, term)
    #       |  (T.PERCENT, factor, term)
    #
    # term: factor (('*'|'/'|'%') factor)*
    def term(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, to_prefix(child))

    #
    # factor --> power
    #         |  (T.PLUS, factor)
    #         |  (T.MINUS, factor)
    #         |  (T.TILDE, factor)
    #
    # factor: ('+'|'-'|'~') factor | power
    def factor(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            return (tok, strip(child))

    #
    # power --> postop
    #        |  (T.DOUBLESTAR, postop, factor)
    #
    # postop --> atom
    #         |  (CALL, atom, arglist)
    #         |  (CALL, atom, ())
    #         |  (NT.subscript, atom, subscript+)
    #         |  (T.DOT, atom, name+)
    #
    # power: atom trailer* ('**' factor)*
    def power(g, tok, where, child):
        if len(child) == 1:
            return (tok, child[0][1])
        l = len(child)
        t1 = child[0][1]
        lop=''
        i = 1
        while i < l:
            ch = child[i]
            if ch[0] is g.T.DOUBLESTAR:
                op = ch[0]
                t2 = child[i+1][1]
                t1 = (op, t1, t2)
                lop = op
                i = i + 2
            elif ch[1] is g.T.DOT:
                op = ch[1]
                if lop == op:
                    t1 = t1 + (ch[2],)
                else:
                    t1 = (op, t1, ch[2])
                    lop = op
                i = i + 1
            else:
                op = ch[1]
                t1 = (op, t1) + ch[2]
                i = i + 1
                lop = op
        return (tok, t1)

    #
    # atom --> test
    #       |  (T.NAME, name)
    #       |  (T.NUMBER, number)
    #       |  (T.STRING, string+)
    #       |  (T.LPAR, test*)
    #       |  (T.LSQB, test*)
    #       |  ('`', test*)
    #       |  (T.LBRACE, (T.COLON, test, test)*)
    #
    #    Note that
    #          (a+b)
    #    will parse to
    #          (T.PLUS, (T.NAME, 'a'), (T.NAME 'b'))
    #    unless it is written as
    #          (a+b,)
    #    in which case it parses to
    #          (T.LPAR (T.PLUS, (T.NAME, 'a'), (T.NAME 'b')), T.COMMA)
    #    which implies that is is a tuple.  This magic happens here.
    #    The trailing comma is left intact for consistency with the other
    #    uses of testlist. (XXX perhaps it should be stripped everwhere?)
    #    
    #
    # atom: '(' [testlist] ')' | '[' [testlist] ']'
    #       | '{' [dictmaker] '}' | '`' testlist '`'
    #       | NAME | NUMBER | STRING+
    def atom(g, tok, where, child):
        if child[0][0] is g.T.LPAR:
            if len(child) == 2:
                return (tok, (child[0][0],))
            else:
                tests = child[1][1][1:]
                if len(tests) == 1:
                    return (tok, tests[0])
                else:
                    return (tok, (g.T.LPAR,) + tests)
        elif child[0][0] is g.T.STRING:
            return (tok, (child[0][0],) + strip(child))
        elif len(child) == 1:
            return (tok, child[0])
        elif len(child) == 2:
            return (tok, (child[0][0],))
        else:
            return (tok, (child[0][0],) + child[1][1][1:])

    #
    # lambdef --> (NT.lambdef, parameters, test)
    #
    # lambdef: 'lambda' [varargslist] ':' test
    def lambdef(g, tok, where, child):
        if len(child) == 4:
            params = (g.NT.parameters,) + child[1][1]
        else:
            params = (g.NT.parameters, )
        return (tok, (tok, params, child[-1][1]))

    # trailer: '(' [arglist] ')' | '[' subscriptlist ']' | '.' NAME
    def trailer(g, tok, where, child):
        op = child[0][0]
        if op is g.T.LPAR:
            op = CALL
        elif op is g.T.LSQB:
            op = g.NT.subscript
        if child[-1][0] is g.T.NAME:
            return (tok, op, (child[-1][1]),)
        elif len(child) == 2:
            return (tok, op, ())
        else:
            return (tok, op) + child[1][1:]

    # subscriptlist: subscript (',' subscript)* [',']
    def subscriptlist(g, tok, where, child):
        return (tok, strip(keep(child, g.NT.subscript)))

    #
    # subscript --> test
    #            |  (ELIPSE,)
    #            |  (NT.sliceop, test)
    #            |  (NT.sliceop, test, test)
    #            |  (NT.sliceop, test, test, test)
    #
    # subscript: '.' '.' '.' | test | [test] ':' [test] [sliceop]
    def subscript(g, tok, where, child):
        if child[0][0] is g.T.DOT:
            return (tok, (ELIPSE,))
        child = strip(keep(child, g.NT.test))
        if len(child) == 1:
            return (tok, (child[0]))
        else:
            return (tok, (g.NT.sliceop,) + child)


    # sliceop: ':' [test]
    def sliceop(g, tok, where, child):
        # cheating here because is is only used by subscript
        return keep(child, g.NT.test)

    #
    # exprlist --> (NT.exprlist, expr+)
    #
    # exprlist: expr (',' expr)* [',']
    def exprlist(g, tok, where, child):
        return (tok, (tok,) + strip(keep(child, g.NT.expr)))

    #
    # testlist --> (NT.testlist, test+)
    #
    # testlist: test (',' test)* [',']
    def testlist(g, tok, where, child):
        if child[-1][0] is g.T.COMMA:
            return (tok, (tok,) + (strip(keep(child, g.NT.test)))+(g.T.COMMA,))
        else:
            return (tok, (tok,) + (strip(keep(child, g.NT.test))))

    # dictmaker: test ':' test (',' test ':' test)* [',']
    def dictmaker(g, tok, where, child):
        r = []
        add = r.append
        i = 0
        nch = len(child)
        while i < nch:
            add(g.T.COLON, child[i][1], child[i+2][1])
            i = i + 4
        return (tok, tuple(r))

    #
    # classdef --> (NT.classdef, name, testlist, suite)
    #
    # classdef: 'class' NAME ['(' testlist ')'] ':' suite
    def classdef(g, tok, where, child):
        child = keep(child, g.T.NAME, g.NT.testlist, g.NT.suite)
        if len(child) == 2:
            return (tok,) + (child[0][1], (g.NT.testlist), child[1][1])
        else:
            return (tok,) + (child[0][1], child[1][1], child[2][1])

    #
    # arglist --> (NT.arglist, argument+)
    #
    # arglist:  argument (',' argument)* [',']
    def arglist(g, tok, where, child):
        if len(child) == 1:
            return (tok, (child[0][1],))
        else:
            return (tok, strip(keep(child, g.NT.argument)))

    #
    # argument --> (NT.argument, test)
    #           |  (NT.argument, name, test)
    #
    # argument: [test '='] test	# Really [keyword '='] test
    def argument(g, tok, where, child):
        child = keep(child, g.NT.test)
        if len(child) == 1:
            return (tok, child[0][1])
        else:
            child = strip(child)
            if child[0][0] is not g.T.NAME:
                # keyword can't be an expression
                return (tok, (g.T.EQUAL, KWERR, child[1]))
            else:
                return (tok, (g.T.EQUAL, child[0][1], child[1]))


    def NAME(g, tok, where, child):
        if hasattr(g.KW, child):
            return (getattr(g.KW, child), child)
        else:
            return (tok, child)
        
    #
    # Anything from the tokenizer
    #
    def terminal(g, tok, where, child):
        return (tok, child)

#    
# Module initialization
#

def _Init():
    parser = pgen.parser_from_string(GRAMMAR)
    if not parser:
	return None

    dict = reductions.__dict__
    for i in parser.NT.__members__ + parser.T.__members__:
	try:
	    parser.reduction[i] = dict[i]
	except KeyError:
	    parser.reduction[i] = dict['terminal']

    for i in dir(parser.KW):
        if i[:2] != '__':
            setattr(parser.KW, i + '_', i)
    return parser


grammar = _Init()
NT = grammar.NT
T = grammar.T
KW = grammar.KW

def parse_string(str, start=grammar.NT.single_input):
    return grammar.parse_string(str, start)

def parse_file(file, start=grammar.NT.file_input):
    return grammar.parse_file(file, start)


#
# test
#

if __name__  == "__main__":
    def show_stmt(s, indent=0):
        space = ' ' * (indent*2)
        if type(s) is not type(()) or not len(s):
            print space, str(s)
        else:
            print space, s[0]
            for n in s[1:]:
                show_stmt(n, indent+1)
    s = parse_file(sys.argv[1])
    show_stmt(s)
